from typing import Optional

from django.contrib.auth import get_user_model
import strawberry
from strawberry.schema_directive import Location
from strawberry.types import Info
import strawberry_django

from gqlauth.core.utils import get_user

# project
from .types_ import UserFilter, UserType

USER_MODEL = get_user_model()


@strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
class Sample:
    message: str = "fdsafdsafdsfa"


@strawberry_django.type(model=USER_MODEL, filters=UserFilter)
class UserQueries:
    @strawberry_django.field
    def public_user(self, info: Info) -> Optional[UserType]:
        user = get_user(info)
        if user.is_authenticated:
            return user  # type: ignore
        return None

    @strawberry_django.field
    def me(self, info: Info) -> Optional[UserType]:
        user = get_user(info)
        if not user.is_anonymous:
            return user  # type: ignore
        return None



from abc import ABC, abstractmethod
from typing import Any, List, Optional

from django.contrib.auth import get_user_model
from django.utils.translation import gettext as _
from jwt import PyJWTError
import strawberry
from strawberry.schema_directive import Location
from strawberry.types import Info

from gqlauth.core.exceptions import TokenExpired
from gqlauth.core.types_ import GQLAuthError, GQLAuthErrors
from gqlauth.core.utils import get_user, get_user_with_status
from gqlauth.jwt.types_ import TokenType
from gqlauth.settings import gqlauth_settings as app_settings

USER_MODEL = get_user_model()

__all__ = ["BaseAuthDirective", "IsAuthenticated", "HasPermission", "IsVerified", "TokenRequired"]


class BaseAuthDirective(ABC):
    @abstractmethod
    def resolve_permission(
        self,
        source: Any,
        info: Info,
        args: list,
        kwargs: dict,
    ) -> Optional[GQLAuthError]:
        ...


@strawberry.schema_directive(
    locations=[
        Location.FIELD_DEFINITION,
    ],
    description="This field requires a JWT token, this token will be used to find the user object.",
)
class TokenRequired(BaseAuthDirective):
    def resolve_permission(
        self, source: Any, info: Info, args: list, kwargs: dict
    ) -> Optional[GQLAuthError]:
        token = app_settings.JWT_TOKEN_FINDER(info)
        try:
            token_type = TokenType.from_token(token)
        except PyJWTError:  # raised by python-jwt
            return GQLAuthError(code=GQLAuthErrors.INVALID_TOKEN)

        except TokenExpired:
            return GQLAuthError(code=GQLAuthErrors.EXPIRED_TOKEN)
        user = token_type.get_user_instance()
        info.context.user = user
        return None


@strawberry.schema_directive(
    locations=[
        Location.FIELD_DEFINITION,
    ],
    description="This field requires authentication",
)
class IsAuthenticated(BaseAuthDirective):
    def resolve_permission(self, source: Any, info: Info, args, kwargs):
        user = get_user(info)
        if not user.is_authenticated:
            return GQLAuthError(code=GQLAuthErrors.UNAUTHENTICATED)
        return None


@strawberry.schema_directive(
    locations=[
        Location.FIELD_DEFINITION,
    ],
    description="This field requires the user to be verified",
)
class IsVerified(BaseAuthDirective):
    def resolve_permission(self, source: Any, info: Info, args, kwargs):
        if (user := get_user_with_status(info)) and user.status.verified:
            return None
        return GQLAuthError(code=GQLAuthErrors.NOT_VERIFIED)


@strawberry.schema_directive(
    locations=[
        Location.FIELD_DEFINITION,
    ],
    description="This field requires a certain permissions to be resolved.",
)
class HasPermission(BaseAuthDirective):
    permissions: strawberry.Private[List[str]]

    def resolve_permission(self, source: Any, info: Info, args, kwargs):
        user = get_user(info)
        for permission in self.permissions:
            if not user.has_perm(permission):  # type: ignore
                return GQLAuthError(
                    code=GQLAuthErrors.NO_SUFFICIENT_PERMISSIONS,
                    message=_(
                        f"User {getattr(user, user.USERNAME_FIELD, user.first_name)}, has not sufficient permissions for {info.path.key}"  # type: ignore
                    ),
                )
        return None



from functools import partial
from typing import Any, AsyncGenerator, Callable, List, Union

from asgiref.sync import sync_to_async
from django.contrib.auth import get_user_model
from strawberry import UNSET
from strawberry.field import StrawberryField
from strawberry_django import django_resolver
from strawberry_django.fields.field import StrawberryDjangoField
from strawberry_django.utils import is_async

from gqlauth.core.directives import BaseAuthDirective
from gqlauth.core.types_ import GQLAuthError

__all__ = ["field"]

USER_MODEL = get_user_model()


class GqlAuthField(StrawberryDjangoField):
    def _resolve(self, source, info, args, kwargs) -> Union[GQLAuthError, Any]:
        for directive in self.directives:
            if isinstance(directive, BaseAuthDirective) and (
                error := directive.resolve_permission(source, info, args, kwargs)
            ):
                return error
        return super().get_result(source, info, args, kwargs)

    async def _resolve_subscriptions(
        self, source, info, args, kwargs
    ) -> Union[AsyncGenerator, GQLAuthError]:
        for directive in self.directives:
            if isinstance(directive, BaseAuthDirective) and (
                error := await sync_to_async(directive.resolve_permission)(
                    source, info, args, kwargs
                )
            ):
                yield error
                return
        async for res in super().get_result(source, info, args, kwargs):
            yield res

    def get_result(self, source, info, args, kwargs):
        if self.is_subscription:
            return self._resolve_subscriptions(source, info, args, kwargs)
        elif is_async():
            return sync_to_async(self._resolve)(source, info, args, kwargs)
        return self._resolve(source, info, args, kwargs)


def field(
    resolver=None,
    *,
    name=None,
    field_name=None,
    filters=UNSET,
    default=UNSET,
    directives: List[BaseAuthDirective] = None,
    **kwargs,
) -> Union[StrawberryField, Callable[..., StrawberryField]]:
    field_ = GqlAuthField(
        python_name=None,
        graphql_name=name,
        filters=filters,
        django_name=field_name,
        default=default,
        directives=directives,
        **kwargs,
    )
    if resolver:
        resolver = django_resolver(resolver)
        field_ = field_(resolver)
    return field_


subscription = partial(field, is_subscription=True)



from gqlauth.core.mixins import RelayMixin
from gqlauth.user.resolvers import (
    ArchiveAccountMixin,
    Captcha,
    DeleteAccountMixin,
    ObtainJSONWebTokenMixin,
    PasswordChangeMixin,
    PasswordResetMixin,
    PasswordSetMixin,
    RefreshTokenMixin,
    RegisterMixin,
    ResendActivationEmailMixin,
    RevokeTokenMixin,
    SendPasswordResetEmailMixin,
    UpdateAccountMixin,
    VerifyAccountMixin,
    VerifyTokenMixin,
)

__all__ = [
    "Register",
    "VerifyAccount",
    "ResendActivationEmail",
    "SendPasswordResetEmail",
    "PasswordSet",
    "PasswordReset",
    "ObtainJSONWebToken",
    "ObtainJSONWebToken",
    "DeleteAccount",
    "PasswordChange",
    "UpdateAccount",
    "VerifyToken",
    "RefreshToken",
    "RevokeToken",
    "Captcha",
]


class Register(RegisterMixin, RelayMixin):
    __doc__ = RegisterMixin.__doc__


class VerifyAccount(
    VerifyAccountMixin,
    RelayMixin,
):
    __doc__ = VerifyAccountMixin.__doc__


class ResendActivationEmail(
    ResendActivationEmailMixin,
    RelayMixin,
):
    __doc__ = ResendActivationEmailMixin.__doc__


class SendPasswordResetEmail(SendPasswordResetEmailMixin, RelayMixin):
    __doc__ = SendPasswordResetEmailMixin.__doc__


class PasswordSet(PasswordSetMixin, RelayMixin):
    __doc__ = PasswordSetMixin.__doc__


class PasswordReset(PasswordResetMixin, RelayMixin):
    __doc__ = PasswordResetMixin.__doc__


class ObtainJSONWebToken(ObtainJSONWebTokenMixin, RelayMixin):
    __doc__ = ObtainJSONWebTokenMixin.__doc__


class ArchiveAccount(ArchiveAccountMixin, RelayMixin):
    __doc__ = ArchiveAccountMixin.__doc__


class DeleteAccount(DeleteAccountMixin, RelayMixin):
    __doc__ = DeleteAccountMixin.__doc__


class PasswordChange(PasswordChangeMixin, RelayMixin):
    __doc__ = PasswordChangeMixin.__doc__


class UpdateAccount(UpdateAccountMixin, RelayMixin):
    __doc__ = UpdateAccountMixin.__doc__


class VerifyToken(VerifyTokenMixin, RelayMixin):
    __doc__ = VerifyTokenMixin.__doc__


class RefreshToken(RefreshTokenMixin, RelayMixin):
    __doc__ = RefreshTokenMixin.__doc__


class RevokeToken(RevokeTokenMixin, RelayMixin):
    __doc__ = RevokeTokenMixin.__doc__



from dataclasses import dataclass, field
from datetime import timedelta
from random import SystemRandom
import typing
from typing import Callable, NewType, Optional, Set, Union

from django.conf import settings as django_settings
from django.utils.module_loading import import_string
from strawberry.annotation import StrawberryAnnotation
from strawberry.field import StrawberryField
from strawberry.types import Info

if typing.TYPE_CHECKING:  # pragma: no cover
    from gqlauth.core.utils import UserProto
    from gqlauth.jwt.types_ import TokenType


def default_text_factory():
    return "".join(
        [
            str(node)
            for node in [
                SystemRandom().randint(0, 10) for _ in range(5, SystemRandom().randint(10, 20))
            ]
        ]
    )


T = typing.TypeVar("T")
DjangoSetting = NewType("DjangoSetting", str)


class ImportString(typing.Generic[T]):
    def __init__(self, path: str):
        self.path = path

    def preform_import(self) -> T:
        return import_string(self.path)

    def __call__(self, *args, **kwargs):
        # FIXME: this is not covered. and just used to fool mypy.
        return import_string(self.path)(*args, **kwargs)


username_field = StrawberryField(
    python_name="username", default=None, type_annotation=StrawberryAnnotation(str)
)
password_field = StrawberryField(
    python_name="password", default=None, type_annotation=StrawberryAnnotation(str)
)
first_name_field = StrawberryField(
    python_name="first_name",
    default=None,
    type_annotation=StrawberryAnnotation(Optional[str]),
)
last_name_field = StrawberryField(
    python_name="last_name",
    default=None,
    type_annotation=StrawberryAnnotation(Optional[str]),
)
email_field = StrawberryField(
    python_name="email", default=None, type_annotation=StrawberryAnnotation(str)
)


@dataclass
class GqlAuthSettings:
    ALLOW_LOGIN_NOT_VERIFIED: bool = False
    """
    """
    LOGIN_FIELDS: Set[StrawberryField] = field(
        default_factory=lambda: {
            username_field,
        }
    )
    """
    These fields would be used to authenticate with SD-jwt `authenticate` function.
    This function will call each of our `AUTHENTICATION_BACKENDS`,
    And will return the user from one of them unless `PermissionDenied` was raised.
    You can pass any fields that would be accepted by your backends.

    **Note that `password field` is mandatory and cannot be removed.**
    """
    LOGIN_REQUIRE_CAPTCHA: bool = True
    """
    whether login will require captcha verification.
    """
    REGISTER_MUTATION_FIELDS: Set[StrawberryField] = field(
        default_factory=lambda: {email_field, username_field}
    )
    """
    fields on register, plus password1 and password2,
    can be a dict like UPDATE_MUTATION_fieldS setting
    """
    REGISTER_REQUIRE_CAPTCHA: bool = True
    """
    whether register will require captcha verification.
    """
    # captcha stuff
    #: captcha expiration delta.
    CAPTCHA_EXPIRATION_DELTA: timedelta = timedelta(seconds=120)
    #: max number of attempts for one captcha.
    CAPTCHA_MAX_RETRIES: int = 5
    CAPTCHA_TEXT_FACTORY: Callable = default_text_factory
    """
    A callable with no arguments that returns a string.
    This will be used to generate the captcha image.
    """
    CAPTCHA_TEXT_VALIDATOR: Callable[[str, str], bool] = (
        lambda original, received: original == received
    )
    """
    A callable that will receive the original string vs user input and returns a boolean.
    """
    FORCE_SHOW_CAPTCHA: bool = False
    """
    Whether to show the captcha image after it has been created for debugging purposes.
    """
    CAPTCHA_SAVE_IMAGE: bool = False
    """
    if True, an png representation of the captcha will be saved under
    MEDIA_ROOT/captcha/<datetime>/<uuid>.png
    """
    # optional fields on update account, can be list of fields
    UPDATE_MUTATION_FIELDS: Set[StrawberryField] = field(
        default_factory=lambda: {first_name_field, last_name_field}
    )
    """
    fields on update account mutation.
    """

    # email tokens
    EXPIRATION_ACTIVATION_TOKEN: timedelta = timedelta(days=7)
    EXPIRATION_PASSWORD_RESET_TOKEN: timedelta = timedelta(hours=1)
    EXPIRATION_PASSWORD_SET_TOKEN: timedelta = timedelta(hours=1)
    # email stuff
    EMAIL_FROM: Union[DjangoSetting, Callable[[], str]] = lambda: getattr(
        django_settings, "DEFAULT_FROM_EMAIL", "test@email.com"
    )
    SEND_ACTIVATION_EMAIL: bool = True
    # client: example.com/activate/token
    ACTIVATION_PATH_ON_EMAIL: str = "activate"
    # client: example.com/password-set/token
    PASSWORD_SET_PATH_ON_EMAIL: str = "password-set"
    # client: example.com/password-reset/token
    PASSWORD_RESET_PATH_ON_EMAIL: str = "password-reset"
    # email subjects templates
    EMAIL_SUBJECT_ACTIVATION: str = "email/activation_subject.txt"
    EMAIL_SUBJECT_ACTIVATION_RESEND: str = "email/activation_subject.txt"
    EMAIL_SUBJECT_PASSWORD_SET: str = "email/password_set_subject.txt"
    EMAIL_SUBJECT_PASSWORD_RESET: str = "email/password_reset_subject.txt"
    # email templates
    EMAIL_TEMPLATE_ACTIVATION: str = "email/activation_email.html"
    EMAIL_TEMPLATE_ACTIVATION_RESEND: str = "email/activation_email.html"
    EMAIL_TEMPLATE_PASSWORD_SET: str = "email/password_set_email.html"
    EMAIL_TEMPLATE_PASSWORD_RESET: str = "email/password_reset_email.html"
    EMAIL_TEMPLATE_VARIABLES: dict = field(default_factory=lambda: {})
    # others
    ALLOW_DELETE_ACCOUNT: bool = False
    """
    If True, DeleteAcount mutation will permanently delete the user.
    """
    ALLOW_PASSWORDLESS_REGISTRATION: bool = False
    """
    Whether to allow registration with no password
    """
    SEND_PASSWORD_SET_EMAIL: bool = False
    # JWT stuff
    JWT_SECRET_KEY: Union[DjangoSetting, Callable[[], str]] = lambda: django_settings.SECRET_KEY
    """
    key used to sign the JWT token.
    """
    JWT_ALGORITHM: str = "HS256"
    """
    Algorithm used for signing the token.
    """
    JWT_TIME_FORMAT: str = "%Y-%m-%dT%H:%M:%S.%f"
    """
    A valid 'strftime' string that will be used to encode the token payload.
    """
    JWT_PAYLOAD_HANDLER: ImportString[Callable[["UserProto"], "TokenType"]] = ImportString(
        "gqlauth.jwt.default_hooks.create_token_type"
    )
    """
    A custom function to generate the token datatype, its up to you to encode the token.
    """
    JWT_PAYLOAD_PK: StrawberryField = field(default_factory=lambda: username_field)
    """
    field that will be used to generate the token from a user instance and
    retrieve user based on the decoded token.
    *This filed must be unique in the database*
    """
    JWT_DECODE_HANDLER: ImportString[Callable[[str], "TokenType"]] = ImportString(
        "gqlauth.jwt.default_hooks.decode_jwt"
    )

    JWT_TOKEN_FINDER: ImportString[Callable[[Info], Optional[str]]] = ImportString(
        "gqlauth.jwt.default_hooks.token_finder"
    )
    """
    A hook called by `GqlAuthRootField` to find the token, **remember to strip the "JWT " prefix
    if you override this.**
    """
    JWT_EXPIRATION_DELTA: timedelta = timedelta(minutes=5)
    """
    Timedelta added to `utcnow()` to set the expiration time.

    When this ends you will have to create a new token by logging in
    or using the refresh token.
    """
    # refresh token stuff
    JWT_LONG_RUNNING_REFRESH_TOKEN: bool = True
    """
    Whether to enable refresh tokens to be used as an alternative to login every time
    the token is expired.
    """
    JWT_REFRESH_TOKEN_N_BYTES: int = 20
    """
    Number of bytes for long running refresh token.
    """
    JWT_REFRESH_EXPIRATION_DELTA: timedelta = timedelta(days=7)
    """
    Refresh token expiration time delta.
    """

    def __post_init__(self):
        # if there override the defaults
        if "email" not in {field_.name for field_ in self.REGISTER_MUTATION_FIELDS}:
            self.SEND_ACTIVATION_EMAIL = False



from django.contrib.auth import get_user_model
from strawberry import auto
from strawberry.annotation import StrawberryAnnotation
from strawberry.field import StrawberryField
from strawberry.types import Info
import strawberry_django

from gqlauth.core.utils import inject_fields
from gqlauth.models import UserStatus
from gqlauth.settings import gqlauth_settings

USER_MODEL = get_user_model()
# UPDATE_MUTATION_FIELDS are here because they are most likely to be in the model.
USER_FIELDS = {
    StrawberryField(
        python_name=USER_MODEL._meta.pk.name,  # type: ignore
        default=None,
        type_annotation=StrawberryAnnotation(auto),
    ),
    StrawberryField(
        python_name=USER_MODEL.USERNAME_FIELD,
        default=None,
        type_annotation=StrawberryAnnotation(auto),
    ),
    StrawberryField(
        python_name=USER_MODEL.EMAIL_FIELD, default=None, type_annotation=StrawberryAnnotation(auto)
    ),
}.union(gqlauth_settings.UPDATE_MUTATION_FIELDS)


@strawberry_django.filters.filter(UserStatus)
class UserStatusFilter:
    verified: auto
    archived: auto


@strawberry_django.type(model=UserStatus, filters=UserStatusFilter)
class UserStatusType:
    verified: auto
    archived: auto


@strawberry_django.filters.filter(USER_MODEL)
@inject_fields(USER_FIELDS, annotations_only=True)
class UserFilter:
    logentry_set: auto
    is_superuser: auto
    last_login: auto
    is_staff: auto
    is_active: auto
    date_joined: auto
    status: UserStatusFilter


@strawberry_django.type(model=USER_MODEL, filters=UserFilter)
@inject_fields(USER_FIELDS, annotations_only=True)
class UserType:
    logentry_set: auto
    is_superuser: auto
    last_login: auto
    is_staff: auto
    is_active: auto
    date_joined: auto
    status: UserStatusType

    @strawberry_django.field
    def archived(self, info: Info) -> bool:
        return self.status.archived

    @strawberry_django.field
    def verified(self, info: Info) -> bool:
        return self.status.verified



import json
from typing import TYPE_CHECKING, Optional, cast

from django.conf import settings as django_settings
from django.contrib.auth import get_user_model
from django.contrib.auth.base_user import AbstractBaseUser
import jwt
from strawberry.types import Info

if TYPE_CHECKING:  # pragma: no cover
    from gqlauth.jwt.types_ import TokenType

USER_MODEL = get_user_model()
app_settings = django_settings.GQL_AUTH

JWT_PREFIX = "JWT "


def token_finder(info: Info) -> Optional[str]:
    headers = info.context.request.headers
    if token := headers.get("authorization", None) or headers.get("Authorization", None):
        assert isinstance(token, str)
        return token.strip(JWT_PREFIX)
    return None


def create_token_type(user: AbstractBaseUser) -> "TokenType":
    from gqlauth.jwt.types_ import TokenPayloadType, TokenType

    user_pk = app_settings.JWT_PAYLOAD_PK.python_name
    pk_field = {user_pk: getattr(user, user_pk)}
    payload = TokenPayloadType(
        **pk_field,
    )
    serialized = json.dumps(payload.as_dict(), sort_keys=True, indent=1)
    return TokenType(
        token=str(
            jwt.encode(
                payload={"payload": serialized},
                key=cast(str, app_settings.JWT_SECRET_KEY),
                algorithm=app_settings.JWT_ALGORITHM,
            )
        ),
        payload=payload,
    )


def decode_jwt(token: str) -> "TokenType":
    from gqlauth.jwt.types_ import TokenPayloadType, TokenType

    decoded = json.loads(
        jwt.decode(
            token,
            key=cast(str, app_settings.JWT_SECRET_KEY),
            algorithms=[
                app_settings.JWT_ALGORITHM,
            ],
        )["payload"]
    )
    return TokenType(token=token, payload=TokenPayloadType.from_dict(decoded))



# Generated by Django 4.1.1 on 2022-10-02 10:53

import uuid

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Captcha",
            fields=[
                (
                    "uuid",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False
                    ),
                ),
                ("text", models.CharField(editable=False, max_length=50)),
                ("insert_time", models.DateTimeField(auto_now_add=True)),
                ("tries", models.IntegerField(default=0)),
                (
                    "image",
                    models.ImageField(
                        editable=False,
                        help_text="url for the captcha image",
                        upload_to="captcha/%Y/%m/%d/",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="UserStatus",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("verified", models.BooleanField(default=False)),
                ("archived", models.BooleanField(default=False)),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="status",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="RefreshToken",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("token", models.CharField(editable=False, max_length=255, verbose_name="token")),
                ("created", models.DateTimeField(auto_now_add=True, verbose_name="created")),
                ("revoked", models.DateTimeField(blank=True, null=True, verbose_name="revoked")),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="refresh_tokens",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "refresh token",
                "verbose_name_plural": "refresh tokens",
                "unique_together": {("token", "revoked")},
            },
        ),
    ]



from dataclasses import dataclass
import io
from pathlib import Path
import sys

from PIL.Image import Image
from django.core.files.base import ContentFile

from gqlauth.settings import gqlauth_settings as app_settings

from .create import ImageCaptcha

sys.path.append(str(Path(__file__).parent.parent.parent))
FONTS_PATH = str(Path(__file__).parent.joinpath("fonts"))


@dataclass
class CaptchaInstanceType:
    pil_image: Image
    text: str

    def to_django(self, name: str) -> ContentFile:
        # inspired by https://stackoverflow.com/questions/34140900
        bytes_array = io.BytesIO()
        self.pil_image.save(bytes_array, format="PNG")
        return ContentFile(bytes_array.getvalue(), name + ".png")

    def show(self):
        self.pil_image.show()


def get_image(text):
    image = ImageCaptcha(
        width=300,
        height=150,
        heb_fonts=[FONTS_PATH + "/stam.ttf"],
        fonts=[FONTS_PATH + "/OpenSans-Semibold.ttf"],
    )
    image = image.generate_image(text)
    return image


def generate_text() -> str:
    return app_settings.CAPTCHA_TEXT_FACTORY()


def generate_captcha_text() -> CaptchaInstanceType:
    text = generate_text()
    image = get_image(text)
    return CaptchaInstanceType(pil_image=image, text=text)
